use std::path::PathBuf;

use based::{
    get_pg, page::Shell, request::{assets::DataResponse, respond_html, RawResponse, RequestContext, StringResponse}
};
use maud::{html, PreEscaped};
use rocket::{get, Data, State};

use crate::archive::{PathEntry, WebsiteArchive};

pub async fn render_page(content: PreEscaped<String>, ctx: RequestContext) -> StringResponse {
    based::page::render_page(
        content,
        "Website Archive",
        ctx,
        &Shell::new(
            html! {
                script src="https://cdn.tailwindcss.com" {};
            },
            html! {},
            Some("bg-zinc-950 text-white min-h-screen flex pt-8 justify-center".to_string()),
        ),
    )
    .await
}

#[get("/favicon/<domain>")]
pub async fn favicon_route(domain: &str) -> Option<DataResponse> {
    let fav: Option<(Vec<u8>,)> = sqlx::query_as("SELECT favicon FROM site_favicon WHERE domain = $1")
    .bind(domain)
    .fetch_optional(get_pg!()).await.unwrap();

    if let Some(fav_data) = fav {
        return Some(DataResponse::new(fav_data.0, "image/x-icon", Some(60 * 60 * 24 * 5)));
    }

    None
}

#[get("/")]
pub async fn index(ctx: RequestContext, arc: &State<WebsiteArchive>) -> StringResponse {
    let websites = arc.domains();

    let content = html! {
        div class="container mx-auto p-4" {
            h1 class="text-5xl font-bold text-center mb-10" { "Websites" };
            div class="grid grid-cols-2 sm:grid-cols-3 lg:grid-cols-4 gap-6 w-screen" {

                @for site in websites {
                    a href=(format!("/d/{site}")) class="bg-neutral-900 shadow-md rounded-lg hover:bg-neutral-800 bg-gray-1 hover:cursor-pointer transition-all duration-300 flex flex-col items-center justify-center aspect-square max-w-60" {
                        div class="bg-blue-500 text-white rounded-full p-4" {
                            img class="h-8 w-8" src=(format!("/favicon/{site}")) {};
                        };
                        p class="mt-4 text-base font-medium" { (site) };
                    };
                };
            };
        }
    };

    render_page(content, ctx).await
}

pub fn arrow_icon(color: &str) -> PreEscaped<String> {
    html! {
        svg class=(format!("w-5 h-5 text-{color}-500")) xmlns="http://www.w3.org/2000/svg" fill="none" viewBox="0 0 24 24" stroke="currentColor" {
            path stroke-linecap="round" stroke-linejoin="round" stroke-width="2" d="M9 5l7 7-7 7" {};
        };
    }
}

#[get("/d/<domain>/<paths..>")]
pub async fn domain_info_route(
    ctx: RequestContext,
    domain: &str,
    paths: PathBuf,
    arc: &State<WebsiteArchive>,
) -> StringResponse {
    let domain = arc.get_domain(domain);
    let paths = domain.paths(paths.to_str().unwrap());

    let content = html! {
        h2 class="text-xl font-bold mb-4" { (domain.name) };
        div class="max-w-md mx-auto p-4 bg-neutral-900 rounded-lg shadow-md" {
            ul class="space-y-2" {
                @for path in paths {
                    a href=(path.url()) class="flex items-center gap-2 p-3 border bg-neutral-800 rounded hover:shadow-lg transition" {
                        @if matches!(path, PathEntry::Document(_)) {
                            (arrow_icon("red"))
                        } @else {
                            (arrow_icon("blue"))
                        };

                        span class="font-medium" { (path.path()) };
                    };
                };
            };
        };
    };

    render_page(content, ctx).await
}

#[get("/s/<domain>/<path..>?<time>")]
pub async fn render_website(
    domain: &str,
    path: PathBuf,
    time: Option<&str>,
    arc: &State<WebsiteArchive>,
) -> Option<StringResponse> {
    let document = arc.get_domain(domain).path(path.to_str().unwrap());

    let content = document.render_local(if time.is_some() {
        Some(time.unwrap().to_string())
    } else {
        None
    });

    if let Some(content) = content {
        return Some(respond_html(&content));
    } else {
        if std::env::var("DOWNLOAD_ON_DEMAND")
            .unwrap_or("false".to_string())
            .as_str()
            == "true"
        {
            arc.archive_url(&format!("https://{domain}/{}", path.to_str().unwrap()));

            return Some(respond_html(&document.render_local(if time.is_some() {
                Some(time.unwrap().to_string())
            } else {
                None
            })?));
        }
    }

    None
}
